#ifndef JENGINE_H
#define JENGINE_H

#include <string>
#include <vector>
#include <set>

#include "util.h"
#include "jarray.h"

/** Interface to dynamically loaded J engine. */
class jengine {
 public:
  friend class jarray;
  /** Pointer to function, implementing monadic variant of a verb. */
  typedef jarray (*monad) (jarray y);
  /** Pointer to function, implementing dyadic variant of a verb. */
  typedef jarray (*dyad) (jarray x, jarray y);
  /** Pointer to function, implementing monadic variant of an adverb. */
  typedef jarray (*amonad) (monad um, dyad ud, jarray y);
  /** Pointer to function, implementing dyadic variant of an adverb. */
  typedef jarray (*adyad) (monad um, dyad ud, jarray x, jarray y);

  /** Initializes J engine. */
  jengine();
  /** Frees J engine memory. */
  ~jengine();

  /** Loads and links J dynamic library, must be called once before this class
   * is instantiated. */
  static void initJlibrary(std::ostream &);
  
  /** Executes J sentence.
   * @param s sentence to execute, must be single line.
   * @return 0 upon success. */
  bool doJ(const std::string s);

  /** Retrieves a named array from J.
   * @param name name of the array to retrieve
   * @return the J array with the specified name, invalid if name not defined. */
  const jarray get(const std::string name);

  /** Assigns J name to the specified array.
   * @param name is the name for the array (may already be defined).
   * @param value is the array to be known under the specified name.
   * @return true if success. */
  bool set(const std::string name, jarray &value);

  /** J error code for the last unsuccessful operation.
   * @return J error code or 0 if no error */ 
  int getError();
  
  /** Conveniently check that there was no error.
   * @return true if there was no error. */
  inline bool ok() {
    return (getError()==0);
  };

  /** Constant to denote the infinite rank. */
  const static int RMAX=10000;

  /** Defines J verb, calling one (or both) of specified C++ functions.
   * Either of mf or df can be NULL, meaning the absence of corresponding 
   * (monadic/dyadic) case. Ranks, equal to RMAX denote "infinite" rank.
   * @param name name of the new verb (may already be defined).
   * @param mf function to be called when verb is invoked monadically.
   * @param df function to be called when verb is invoked dyadically.
   * @param mr monadic rank (infinite by default).
   * @param lr left rank (infinite by default).
   * @param rr right rank (infinite by default).
   * @return "true" on success. */
  bool defVerb(std::string name, monad mf, dyad df, int mr=RMAX, int lr=RMAX, int rr=RMAX);

  /** defines J adverb, calling one (or both) of specified C++ functions.
   * Works similarly to verb definition, except the functions may now call
   * the verb argument of an adverb monadically or dyadically.
   * Adverbs always have infinite ranks.
   * @param name name of the new adverb (may already be defined).
   * @param mf function to be called when adverb is invoked monadically.
   * @param df function to be called when adverb is invoked dyadically.
   * @return "true" on success. */
  bool defAdverb(std::string name, amonad mf, adyad df);

  /** Defines a J script, given by a (possibly multi-libe) string.
   * @param name name of the script.
   * @param type type of the script (the left argument of ":" conjunction).
   * @param code code of the script (possibly multi-line).
   * @param mr monadic rank (infinite by default).
   * @param lr left rank (infinite by default).
   * @param rr right rank (infinite by default). 
   * @return "true" upon success. */
  bool defScript(std::string name, int type, std::string code, 
		 int mr=jengine::RMAX, 
		 int lr=jengine::RMAX, 
		 int rr=jengine::RMAX);

  /** Checks if given name was defined via defScript.
   * @param name name to check 
   * @return true if the name was defined via defScript */
  bool isBuiltin(std::string name) const;

  /** Returns set of all names defined via defScript.
   * @return set of all names, defined via defScript */
  std::set<std::string> getBuiltins() const;

 protected:
  /** Allocates J array.
   * @param t type of the array.
   * @param n number of elements in the array.
   * @param r rank of the array.
   * @param s shape of the array.
   * @return pointer to header of newly allocated array. */
  static void* GA(const jarray::I t, const jarray::I n, const jarray::I r, const jarray::I *s);
  /** Frees J array memory.
   * @param hdr pointer to the array header.
   */
  static void FR(void *hdr);
  
  /** Checks if J engine is currently initialized.
   * @return "true" if J engine is initialized. */
  static bool isInitialized();

 private:
  // copy disabled
  jengine(const jengine &);
  const jengine& operator=(const jengine&);
  
  class Impl;
  Impl* pImpl;
};
#endif
