NB.---------------------------------------------------------------
NB. Quadvav 1-d numerical integrator version 1.0
NB. Based on quadva integrator by L.F. Shampine, but additionally
NB. can integrate arbitrarily shaped vectors of functions. Handles
NB. mild singularities at endpoints and infinite intervals.
NB.
NB. (c) 2007 Konstantin L. Metlov <metlov@kinetic.ac.donetsk.ua>
NB. All Rights Reserved
NB.
NB. Permission is granted to use, modify and distribute for 
NB. non-commercial research and testing purposes leaving this
NB. copyright notice intact.

NB. ***************************
NB.
NB. USAGE: [x] u quadvav y
NB.
NB. quadvav is an adverb, whose left argument is the integrand.
NB.
NB. When applied to a list of numbers (say, of length N), the
NB. integrand must return an array with the last dimension equal
NB. to N, but otherwise of an arbitrary (but constant across
NB. different calls) shape. If S is the shape of this array minus
NB. the last dimension, the integration will produce an array of
NB. shape "2, S" with the values of integrals in the first element,
NB. and the errors in the last. Negative error means that the
NB. corresponding component of the integral did not converge to
NB. requested accuracy.
NB.
NB. x, if specified, must be a list of two numbers 'atol, rtol'.
NB. Refinement stops when 'error <: atol >. | rtol * I', where I
NB. is the value of the integral. By default x is '1e_10 1e_5'.
NB.
NB. y specifies integration limits. It can be a sorted list of an
NB. arbitrary length, whose intermediate elements specify the ini-
NB. tial subdivision breakpoints (useful when integrand has known
NB. breaks or peaks). Complex values of y mean that the integration
NB. will be performed along a polygonal contour in the complex
NB. plane (in this case integrand is not allowed to have singula-
NB. rities on the contour).
NB.
NB. EXAMPLES:
NB.
NB.    ([: % *:) quadvav 1 _
NB. 1 1.47143e_8
NB.    ([: %@%: 1 - *:) quadvav _1 1
NB. 3.14159 9.3627e_9
NB.    ((i. 5)&(^~"0 1)) quadvav 0 1
NB.           1         0.5    0.333333        0.25        0.2
NB. 4.96565e_17 3.52687e_17 3.80486e_17 2.59796e_17 5.43357e_7
NB.    (^ @ - @ *:) quadvav __ _
NB. 1.77245 1.2648e_5
NB.
NB. ************************************************************


NB. Integration engine. Makes no substitutions (and so is
NB. completely ignorant of singularities and infinite intervals).
vadapt=: 2 : 0
[:
:
'atol rtol' =. x
if. rtol<:0 do. rtol=. 0 else. rtol =. rtol >. (9!:18) '' end.
atol =. atol >. 0
if. 0 = atol + rtol do. atol =. *: rtol =. 1e_5 end.

NB. Gauss-Kronrod (7,15) pair.
n  =.    0.2077849550078985 0.4058451513773972 0.5860872354676911
n  =. n, 0.7415311855993944 0.8648644233597691 0.9491079123427585
n  =. n, 0.9914553711208126
n  =. (- |. n), 0, n
w  =.    0.2044329400752989 0.1903505780647854 0.1690047266392679
w  =. w, 0.1406532597155259 0.1047900103222502 0.06309209262997855
w  =. w, 0.02293532201052922
w  =. (|. w), 0.2094821410847278, w
w7 =. 0 0.3818300505051189 0 0.2797053914892767
w7 =. w7, 0 0.1294849661688697 0
w  =. w ,: w - (|. w7), 0.4179591836734694, w7

subs =. (}: ,: }.) y    NB. init sub-intervals
Iok =. 0                NB. init accumulator
while. 1 do.
  tt =. (-: +/ subs) + n *"_ 0 (halfh =. - -: -/ subs)
  ar =. v (tr =. , tt)
  if. ([: +./ 2 =/\ ]) ar do. break. end.    NB. args too close
  fx =. (, $~ ($tt) ,~ [: }: $) tr u ar
  Isubs =. halfh *"1 +/"1 w (*"1)/ fx
  I =. (_1}~ |@{:) Iok + +/"1 Isubs
  tol =. atol >. | ({. I) * rtol
  if. *./ , ({: I) <:"0 tol do. break. end.  NB. converged
  NB. locate intervals, where evaluation is precise enough
  okint =. *./"1 ,. (halfh *"0 _ tol) >:"_1 | |: {: Isubs
  if. +./ okint do.
     Iok =. Iok + +/"1 okint *"1 Isubs
     subs =. (-. okint) #"1 subs
     if. (0 = # |: subs) do. break. end.     NB. no more work
  end.
  if. 650 < +: {: $ subs do. break. end.     NB. too much work
  subs =. ({. ,. |.@{:)@(,:"1 -:@+/) subs    NB. subdivide
end.
(_1}~ (* tol <:@+:@> ])@{:) I    NB. mark and return
)

quadvav=: 1 : 0
1e_10 1e_5 u quadvav y
:
'a b' =. 0 _1 { y
if. 0 ~: +/ | {:"1 +. y do. x ((u@:]) vadapt ]) y return. end.
if. a = b do. ,:~ ($ u a) $ 0 return. end.
if. a > b do. (-@{. ,: {:) x u quadvav |. y return. end.
if. +./ 2 >:/\ y do. 'interval must be monotonic' return. end.

NB. mappings and jacobians for various interval types
'`sin asin ash th' =. (1&o.)`(_1&o.)`(_5&o.)`(7&o.)  NB. trigs 
hom =. [: *:@th [: %:@ash %:     NB. half-open interval mapping
select. #. _ = | a, b
case. 0 do.                      NB. mapping of [-1,1] to [a,b]
 'apb bma' =. (-: a + b), (-: -: b - a)
 tx =. apb + bma * ] * 3 - *:
 xt =. ([: +:@sin 3 %~ asin)@(bma %~ [: -: -&apb)
 J  =. (3 * bma) * 1 - *:
case. 1 do.                      NB. mapping of [0,1) to [a,_)
 tx =. a + [: *: ] % <:
 xt =. [: hom -&a
 J  =. _2 * ] % [: (* *:) <:
case. 2 do.                      NB. mapping of (_1,0] to (__,b]
 tx =. b - [: *: ] % >:
 xt =. [: -@hom b&-
 J  =. _2 * ] % [: (* *:) >:
case. 3 do.                      NB. mapping of (_1,1) to (__,_)
 tx =. - % <:@*:
 xt =. th @ -: @ ash @ +:
 J  =. (>: % *:@<:) @ *:
end.

x (((J f.)@:[ *"1 u@:]) vadapt (tx f.)) xt y

NB. 'J f.' can be '(tx d. 1) f.'
)
